from __future__ import annotations

from asyncio import Future
from collections.abc import Callable
from dataclasses import dataclass
from typing import Any

from omu.client import Client
from omu.extension import Extension, ExtensionType
from omu.helper import Coro
from omu.identifier import Identifier
from omu.network.packet import PacketType

from .endpoint import EndpointType
from .packets import (
    EndpointDataPacket,
    EndpointErrorPacket,
    EndpointRegisterPacket,
)

ENDPOINT_EXTENSION_TYPE = ExtensionType(
    "endpoint",
    lambda client: EndpointExtension(client),
    lambda: [],
)


@dataclass(frozen=True, slots=True)
class EndpointHandler:
    endpoint_type: EndpointType
    func: Coro[[Any], Any]


class EndpointExtension(Extension):
    def __init__(self, client: Client) -> None:
        self.client = client
        self.response_futures: dict[int, Future] = {}
        self.registered_endpoints: dict[Identifier, EndpointHandler] = {}
        self.call_id = 0
        client.network.register_packet(
            ENDPOINT_REGISTER_PACKET,
            ENDPOINT_CALL_PACKET,
            ENDPOINT_RECEIVE_PACKET,
            ENDPOINT_ERROR_PACKET,
        )
        client.network.add_packet_handler(ENDPOINT_RECEIVE_PACKET, self._on_receive)
        client.network.add_packet_handler(ENDPOINT_ERROR_PACKET, self._on_error)
        client.network.add_packet_handler(ENDPOINT_CALL_PACKET, self._on_call)
        client.network.add_task(self.on_ready)

    async def _on_receive(self, packet: EndpointDataPacket) -> None:
        if packet.key not in self.response_futures:
            raise Exception(f"Received response for unknown call key {packet.key}")
        future = self.response_futures.pop(packet.key)
        future.set_result(packet.data)

    async def _on_error(self, packet: EndpointErrorPacket) -> None:
        if packet.key not in self.response_futures:
            raise Exception(f"Received error for unknown call key {packet.key}")
        future = self.response_futures.pop(packet.key)
        future.set_exception(Exception(packet.error))

    async def _on_call(self, packet: EndpointDataPacket) -> None:
        if packet.id not in self.registered_endpoints:
            raise Exception(f"Received call for unknown endpoint {packet.id}")
        handler = self.registered_endpoints[packet.id]
        endpoint = handler.endpoint_type
        func = handler.func
        try:
            req = endpoint.request_serializer.deserialize(packet.data)
            res = await func(req)
            res_data = endpoint.response_serializer.serialize(res)
            await self.client.send(
                ENDPOINT_RECEIVE_PACKET,
                EndpointDataPacket(id=packet.id, key=packet.key, data=res_data),
            )
        except Exception as e:
            await self.client.send(
                ENDPOINT_ERROR_PACKET,
                EndpointErrorPacket(id=packet.id, key=packet.key, error=str(e)),
            )
            raise e

    async def on_ready(self) -> None:
        endpoints = {
            key: endpoint.endpoint_type.permission_id
            for key, endpoint in self.registered_endpoints.items()
        }
        packet = EndpointRegisterPacket(endpoints=endpoints)
        await self.client.send(ENDPOINT_REGISTER_PACKET, packet)

    def register[Req, Res](
        self, type: EndpointType[Req, Res], func: Coro[[Req], Res]
    ) -> None:
        if self.client.ready:
            raise Exception("Cannot register endpoint after client is ready")
        if type.id in self.registered_endpoints:
            raise Exception(f"Endpoint for key {type.id} already registered")
        self.registered_endpoints[type.id] = EndpointHandler(
            endpoint_type=type,
            func=func,
        )

    def bind[T, R](
        self,
        handler: Coro[[T], R] | None = None,
        endpoint_type: EndpointType[T, R] | None = None,
    ):
        if endpoint_type is None:
            raise Exception("Endpoint type must be provided")

        def decorator(func: Coro[[T], R]) -> Coro[[T], R]:
            self.register(endpoint_type, func)
            return func

        if handler:
            decorator(handler)
        return decorator

    def listen(
        self, handler: Coro | None = None, name: str | None = None
    ) -> Callable[[Coro], Coro]:
        def decorator(func: Coro) -> Coro:
            type = EndpointType.create_json(
                self.client.app.id,
                name or func.__name__,
            )
            self.register(type, func)
            return func

        if handler:
            decorator(handler)
        return decorator

    async def call[Req, Res](self, endpoint: EndpointType[Req, Res], data: Req) -> Res:
        try:
            self.call_id += 1
            future = Future[bytes]()
            self.response_futures[self.call_id] = future
            json = endpoint.request_serializer.serialize(data)
            await self.client.send(
                ENDPOINT_CALL_PACKET,
                EndpointDataPacket(id=endpoint.id, key=self.call_id, data=json),
            )
            res = await future
            return endpoint.response_serializer.deserialize(res)
        except Exception as e:
            raise Exception(f"Error calling endpoint {endpoint.id.key()}") from e


ENDPOINT_REGISTER_PACKET = PacketType[EndpointRegisterPacket].create_serialized(
    ENDPOINT_EXTENSION_TYPE,
    "register",
    EndpointRegisterPacket,
)
ENDPOINT_CALL_PACKET = PacketType[EndpointDataPacket].create_serialized(
    ENDPOINT_EXTENSION_TYPE,
    "call",
    EndpointDataPacket,
)
ENDPOINT_RECEIVE_PACKET = PacketType[EndpointDataPacket].create_serialized(
    ENDPOINT_EXTENSION_TYPE,
    "receive",
    EndpointDataPacket,
)
ENDPOINT_ERROR_PACKET = PacketType[EndpointErrorPacket].create_serialized(
    ENDPOINT_EXTENSION_TYPE,
    "error",
    EndpointErrorPacket,
)



from __future__ import annotations

import abc
import asyncio
from typing import TYPE_CHECKING

from omu.app import App
from omu.event_emitter import EventEmitter, Unlisten
from omu.helper import Coro

if TYPE_CHECKING:
    from omu.extension import ExtensionRegistry
    from omu.extension.dashboard import DashboardExtension
    from omu.extension.endpoint import EndpointExtension
    from omu.extension.i18n import I18nExtension
    from omu.extension.logger import LoggerExtension
    from omu.extension.permission import PermissionExtension
    from omu.extension.plugin import PluginExtension
    from omu.extension.registry import RegistryExtension
    from omu.extension.server import ServerExtension
    from omu.extension.signal import SignalExtension
    from omu.extension.table import TableExtension
    from omu.network import Network
    from omu.network.packet import PacketType


class ClientEvents:
    def __init__(self) -> None:
        self.started = EventEmitter[[]]()
        self.stopped = EventEmitter[[]]()
        self.ready = EventEmitter[[]]()


class Client(abc.ABC):
    @property
    @abc.abstractmethod
    def ready(self) -> bool: ...

    @property
    @abc.abstractmethod
    def app(self) -> App: ...

    @property
    @abc.abstractmethod
    def loop(self) -> asyncio.AbstractEventLoop: ...

    @property
    @abc.abstractmethod
    def network(self) -> Network: ...

    @property
    @abc.abstractmethod
    def extensions(self) -> ExtensionRegistry: ...

    @property
    @abc.abstractmethod
    def endpoints(self) -> EndpointExtension: ...

    @property
    @abc.abstractmethod
    def plugins(self) -> PluginExtension: ...

    @property
    @abc.abstractmethod
    def tables(self) -> TableExtension: ...

    @property
    @abc.abstractmethod
    def registry(self) -> RegistryExtension: ...

    @property
    @abc.abstractmethod
    def signal(self) -> SignalExtension: ...

    @property
    @abc.abstractmethod
    def server(self) -> ServerExtension: ...

    @property
    @abc.abstractmethod
    def permissions(self) -> PermissionExtension: ...

    @property
    @abc.abstractmethod
    def dashboard(self) -> DashboardExtension: ...

    @property
    @abc.abstractmethod
    def i18n(self) -> I18nExtension: ...

    @property
    @abc.abstractmethod
    def logger(self) -> LoggerExtension: ...

    @property
    @abc.abstractmethod
    def running(self) -> bool: ...

    @abc.abstractmethod
    def run(self) -> None: ...

    @abc.abstractmethod
    async def start(self) -> None: ...

    @abc.abstractmethod
    async def stop(self) -> None: ...

    @abc.abstractmethod
    async def send[T](self, type: PacketType[T], data: T) -> None: ...

    @property
    @abc.abstractmethod
    def event(self) -> ClientEvents: ...

    @abc.abstractmethod
    def on_ready(self, coro: Coro[[], None]) -> Unlisten: ...



from __future__ import annotations

import re
import urllib.parse
from pathlib import Path
from typing import Final

from omu.helper import generate_md5_hash, sanitize_filename
from omu.model import Model

from .interface import Keyable

NAMESPACE_REGEX = re.compile(r"^(\.[^/:.]|[\w-])+$")
NAME_REGEX = re.compile(r"^[^/:.]+$")


class Identifier(Model[str], Keyable):
    def __init__(self, namespace: str, *path: str) -> None:
        self.validate(namespace, *path)
        self.namespace: Final[str] = namespace
        self.path: Final[tuple[str, ...]] = path

    @classmethod
    def validate(cls, namespace: str, *path: str) -> None:
        if not namespace:
            raise Exception("Invalid namespace: Namespace cannot be empty")
        if len(path) == 0:
            raise Exception("Invalid path: Path must have at least one name")
        if not NAMESPACE_REGEX.match(namespace):
            raise Exception(
                f"Invalid namespace: Namespace must match {NAMESPACE_REGEX.pattern}"
            )
        for name in path:
            if not NAME_REGEX.match(name):
                raise Exception(f"Invalid name: Name must match {NAME_REGEX.pattern}")

    @classmethod
    def format(cls, namespace: str, *path: str) -> str:
        cls.validate(namespace, *path)
        return f"{namespace}:{'/'.join(path)}"

    @classmethod
    def from_key(cls, key: str) -> Identifier:
        separator = key.find(":")
        if separator == -1:
            raise Exception(f"Invalid key: No separator found in {key}")
        if key.find(":", separator + 1) != -1:
            raise Exception(f"Invalid key: Multiple separators found in {key}")
        namespace, path = key[:separator], key[separator + 1 :]
        if not namespace or not path:
            raise Exception("Invalid key: Namespace and path cannot be empty")
        return cls(namespace, *path.split("/"))

    @classmethod
    def from_url(cls, url: str) -> Identifier:
        parsed = urllib.parse.urlparse(url)
        namespace = cls.namespace_from_url(url)
        path = parsed.path.split("/")[1:]
        return cls(namespace, *path)

    @classmethod
    def namespace_from_url(cls, url: str) -> str:
        parsed = urllib.parse.urlparse(url)
        return ".".join(reversed(parsed.netloc.split(".")))

    def to_json(self) -> str:
        return self.key()

    @classmethod
    def from_json(cls, json: str) -> Identifier:
        return cls.from_key(json)

    def key(self) -> str:
        return self.format(self.namespace, *self.path)

    def get_sanitized_path(self) -> Path:
        namespace = (
            f"{sanitize_filename(self.namespace)}-{generate_md5_hash(self.namespace)}"
        )
        return Path(namespace, *self.path)

    def is_subpath_of(self, base: Identifier) -> bool:
        return (
            self.namespace == base.namespace
            and self.path[: len(base.path)] == base.path
        )

    def is_namepath_equal(
        self,
        other: Identifier,
        path_length: int | None = None,
    ) -> bool:
        if path_length is None:
            path_length = len(self.path)
        return (
            self.namespace == other.namespace
            and self.path[:path_length] == other.path[:path_length]
        )

    def join(self, *path: str) -> Identifier:
        return Identifier(self.namespace, *self.path, *path)

    def __truediv__(self, name: str) -> Identifier:
        return self.join(name)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Identifier):
            return NotImplemented
        return self.key() == other.key()

    def __hash__(self) -> int:
        return hash(self.key())

    def __repr__(self) -> str:
        return f"Identifier({self.key()})"

    def __str__(self) -> str:
        return self.key()



from __future__ import annotations

from typing import TypedDict

from omu.app import App
from omu.client import Client
from omu.extension import Extension, ExtensionType
from omu.extension.endpoint import EndpointType
from omu.extension.table import TablePermissions, TableType
from omu.identifier import Identifier
from omu.network.packet import PacketType
from omu.serializer import Serializer

from .packets import (
    PermissionRequestPacket,
    PluginRequestPacket,
)

DASHBOARD_EXTENSION_TYPE = ExtensionType(
    "dashboard",
    lambda client: DashboardExtension(client),
    lambda: [],
)


class DashboardSetResponse(TypedDict):
    success: bool


DASHBOARD_SET_PERMISSION_ID = DASHBOARD_EXTENSION_TYPE / "set"
DASHBOARD_SET_ENDPOINT = EndpointType[Identifier, DashboardSetResponse].create_json(
    DASHBOARD_EXTENSION_TYPE,
    "set",
    request_serializer=Serializer.model(Identifier),
    permission_id=DASHBOARD_SET_PERMISSION_ID,
)
DASHBOARD_PERMISSION_REQUEST_PACKET = PacketType[
    PermissionRequestPacket
].create_serialized(
    DASHBOARD_EXTENSION_TYPE,
    "permission_request",
    serializer=PermissionRequestPacket,
)
DASHBOARD_PERMISSION_ACCEPT_PACKET = PacketType[str].create_json(
    DASHBOARD_EXTENSION_TYPE,
    "permission_accept",
)
DASHBOARD_PERMISSION_DENY_PACKET = PacketType[str].create_json(
    DASHBOARD_EXTENSION_TYPE,
    "permission_deny",
)
DASHBOARD_PLUGIN_REQUEST_PACKET = PacketType[PluginRequestPacket].create_serialized(
    DASHBOARD_EXTENSION_TYPE,
    "plugin_request",
    serializer=PluginRequestPacket,
)
DASHBOARD_PLUGIN_ACCEPT_PACKET = PacketType[str].create_json(
    DASHBOARD_EXTENSION_TYPE,
    "plugin_accept",
)
DASHBOARD_PLUGIN_DENY_PACKET = PacketType[str].create_json(
    DASHBOARD_EXTENSION_TYPE,
    "plugin_deny",
)
DASHBOARD_OPEN_APP_PERMISSION_ID = DASHBOARD_EXTENSION_TYPE / "app" / "open"
DASHBOARD_OPEN_APP_ENDPOINT = EndpointType[App, None].create_json(
    DASHBOARD_EXTENSION_TYPE,
    "open_app",
    request_serializer=Serializer.model(App),
    permission_id=DASHBOARD_OPEN_APP_PERMISSION_ID,
)
DASHBOARD_OPEN_APP_PACKET = PacketType[App].create_json(
    DASHBOARD_EXTENSION_TYPE,
    "open_app",
    Serializer.model(App),
)
DASHOBARD_APP_READ_PERMISSION_ID = DASHBOARD_EXTENSION_TYPE / "app" / "read"
DASHOBARD_APP_EDIT_PERMISSION_ID = DASHBOARD_EXTENSION_TYPE / "app" / "edit"
DASHBOARD_APP_TABLE_TYPE = TableType.create_model(
    DASHBOARD_EXTENSION_TYPE,
    "apps",
    App,
    permissions=TablePermissions(
        read=DASHOBARD_APP_READ_PERMISSION_ID,
        write=DASHOBARD_APP_EDIT_PERMISSION_ID,
        remove=DASHOBARD_APP_EDIT_PERMISSION_ID,
    ),
)


class DashboardExtension(Extension):
    def __init__(self, client: Client):
        self.client = client
        self.client.network.register_packet(
            DASHBOARD_PERMISSION_REQUEST_PACKET,
            DASHBOARD_PERMISSION_ACCEPT_PACKET,
            DASHBOARD_PERMISSION_DENY_PACKET,
            DASHBOARD_PLUGIN_REQUEST_PACKET,
            DASHBOARD_PLUGIN_ACCEPT_PACKET,
            DASHBOARD_PLUGIN_DENY_PACKET,
            DASHBOARD_OPEN_APP_PACKET,
        )
        self.apps = client.tables.get(DASHBOARD_APP_TABLE_TYPE)

    async def open_app(self, app: App) -> None:
        if not self.client.permissions.has(DASHBOARD_OPEN_APP_PERMISSION_ID):
            error = f"Pemission {DASHBOARD_OPEN_APP_PERMISSION_ID} required to open app {app}"
            raise PermissionError(error)
        await self.client.endpoints.call(DASHBOARD_OPEN_APP_ENDPOINT, app)



from __future__ import annotations

import asyncio
from dataclasses import dataclass
from typing import Literal

from loguru import logger

from omu.address import Address
from omu.client import Client
from omu.errors import (
    AnotherConnection,
    InvalidOrigin,
    InvalidPacket,
    InvalidToken,
    InvalidVersion,
    OmuError,
    PermissionDenied,
)
from omu.event_emitter import EventEmitter
from omu.helper import Coro
from omu.identifier import Identifier
from omu.token import TokenProvider

from .connection import Connection
from .packet import Packet, PacketType
from .packet.packet_types import (
    PACKET_TYPES,
    ConnectPacket,
    DisconnectPacket,
    DisconnectType,
)
from .packet_mapper import PacketMapper


@dataclass(frozen=True, slots=True)
class PacketHandler[T]:
    packet_type: PacketType[T]
    event: EventEmitter[T]


class Network:
    def __init__(
        self,
        client: Client,
        address: Address,
        token_provider: TokenProvider,
        connection: Connection,
    ):
        self._client = client
        self._address = address
        self._token_provider = token_provider
        self._connection = connection
        self._connected = False
        self._closed = False
        self._event = NetworkEvents()
        self._tasks: list[Coro[[], None]] = []
        self._packet_mapper = PacketMapper()
        self._packet_handlers: dict[Identifier, PacketHandler] = {}
        self.register_packet(
            PACKET_TYPES.CONNECT,
            PACKET_TYPES.DISCONNECT,
            PACKET_TYPES.TOKEN,
            PACKET_TYPES.READY,
        )
        self.add_packet_handler(PACKET_TYPES.TOKEN, self.handle_token)
        self.add_packet_handler(PACKET_TYPES.DISCONNECT, self.handle_disconnect)
        self.add_packet_handler(PACKET_TYPES.READY, self.handle_ready)

    async def handle_token(self, token: str | None):
        if token is None:
            return
        self._token_provider.store(self._address, self._client.app, token)

    async def handle_disconnect(self, reason: DisconnectPacket):
        if reason.type in {
            DisconnectType.SHUTDOWN,
            DisconnectType.CLOSE,
        }:
            return

        self._closed = True
        ERROR_MAP: dict[DisconnectType, type[OmuError]] = {
            DisconnectType.ANOTHER_CONNECTION: AnotherConnection,
            DisconnectType.PERMISSION_DENIED: PermissionDenied,
            DisconnectType.INVALID_TOKEN: InvalidToken,
            DisconnectType.INVALID_ORIGIN: InvalidOrigin,
            DisconnectType.INVALID_VERSION: InvalidVersion,
            DisconnectType.INVALID_PACKET: InvalidPacket,
            DisconnectType.INVALID_PACKET_TYPE: InvalidPacket,
            DisconnectType.INVALID_PACKET_DATA: InvalidPacket,
        }
        error = ERROR_MAP.get(reason.type)
        if error:
            raise error(reason.message)

    async def handle_ready(self, _: None):
        await self._client.event.ready.emit()

    @property
    def address(self) -> Address:
        return self._address

    def set_connection(self, connection: Connection) -> None:
        if self._connected:
            raise RuntimeError("Cannot change connection while connected")
        if self._connection:
            del self._connection
        self._connection = connection

    def set_token_provider(self, token_provider: TokenProvider) -> None:
        self._token_provider = token_provider

    def register_packet(self, *packet_types: PacketType) -> None:
        self._packet_mapper.register(*packet_types)
        for packet_type in packet_types:
            if self._packet_handlers.get(packet_type.id):
                raise ValueError(f"Packet type {packet_type.id} already registered")
            self._packet_handlers[packet_type.id] = PacketHandler(
                packet_type,
                EventEmitter(),
            )

    def add_packet_handler[T](
        self,
        packet_type: PacketType[T],
        packet_handler: Coro[[T], None] | None = None,
    ):
        if not self._packet_handlers.get(packet_type.id):
            raise ValueError(f"Packet type {packet_type.id} not registered")

        def decorator(func: Coro[[T], None]) -> None:
            self._packet_handlers[packet_type.id].event.listen(func)

        if packet_handler:
            decorator(packet_handler)
        return decorator

    @property
    def connected(self) -> bool:
        return self._connected

    async def connect(self, *, reconnect: bool = True) -> None:
        if self._connected:
            raise RuntimeError("Already connected")
        if self._closed:
            raise RuntimeError("Connection closed")

        await self.disconnect()
        try:
            await self._connection.connect()
        except Exception as e:
            if reconnect:
                logger.error(e)
                await self.try_reconnect()
                return
            else:
                raise e
        self._connected = True
        await self.send(
            Packet(
                PACKET_TYPES.CONNECT,
                ConnectPacket(
                    app=self._client.app,
                    token=self._token_provider.get(self._address, self._client.app),
                ),
            )
        )
        listen_task = self._client.loop.create_task(self._listen_task())

        await self._event.status.emit("connected")
        await self._event.connected.emit()
        await self._dispatch_tasks()

        await self.send(Packet(PACKET_TYPES.READY, None))
        await listen_task

        if reconnect:
            await self.try_reconnect()

    async def try_reconnect(self) -> None:
        if self._closed:
            return
        await asyncio.sleep(1)
        await self.connect(reconnect=True)

    async def disconnect(self) -> None:
        if self._connection.closed:
            return
        self._connected = False
        await self._connection.close()
        await self._event.status.emit("disconnected")
        await self._event.disconnected.emit()

    async def send(self, packet: Packet) -> None:
        if not self._connected:
            raise RuntimeError("Not connected")
        await self._connection.send(packet, self._packet_mapper)

    async def _listen_task(self):
        while not self._connection.closed:
            packet = await self._connection.receive(self._packet_mapper)
            self._client.loop.create_task(self.dispatch_packet(packet))

    async def dispatch_packet(self, packet: Packet) -> None:
        await self._event.packet.emit(packet)
        packet_handler = self._packet_handlers.get(packet.type.id)
        if not packet_handler:
            return
        await packet_handler.event(packet.data)

    @property
    def event(self) -> NetworkEvents:
        return self._event

    def add_task(self, task: Coro[[], None]) -> None:
        if self._client.ready:
            raise RuntimeError("Cannot add task after client is ready")
        self._tasks.append(task)

    def remove_task(self, task: Coro[[], None]) -> None:
        self._tasks.remove(task)

    async def _dispatch_tasks(self) -> None:
        for task in self._tasks:
            await task()


type NetworkStatus = Literal["connecting", "connected", "disconnected"]


class NetworkEvents:
    def __init__(self) -> None:
        self.connected = EventEmitter[[]]()
        self.disconnected = EventEmitter[[]]()
        self.packet = EventEmitter[Packet]()
        self.status = EventEmitter[NetworkStatus]()



from __future__ import annotations

from enum import Enum
from typing import TypedDict

from omu.app import App, AppJson
from omu.identifier import Identifier
from omu.model import Model
from omu.serializer import Serializer

from .packet import PacketType


class ConnectPacketData(TypedDict):
    app: AppJson
    token: str | None


class ConnectPacket(Model[ConnectPacketData]):
    def __init__(
        self,
        app: App,
        token: str | None = None,
    ):
        self.app = app
        self.token = token

    def to_json(self) -> ConnectPacketData:
        return {
            "app": self.app.to_json(),
            "token": self.token,
        }

    @classmethod
    def from_json(cls, json: ConnectPacketData) -> ConnectPacket:
        return cls(
            app=App.from_json(json["app"]),
            token=json["token"],
        )


class DisconnectType(str, Enum):
    INVALID_TOKEN = "invalid_token"
    INVALID_ORIGIN = "invalid_origin"
    INVALID_VERSION = "invalid_version"
    INVALID_PACKET_TYPE = "invalid_packet_type"
    INVALID_PACKET_DATA = "invalid_packet_data"
    INVALID_PACKET = "invalid_packet"
    ANOTHER_CONNECTION = "another_connection"
    PERMISSION_DENIED = "permission_denied"
    SHUTDOWN = "shutdown"
    CLOSE = "close"


class DisconnectPacketData(TypedDict):
    type: str
    message: str | None


class DisconnectPacket(Model[DisconnectPacketData]):
    def __init__(self, type: DisconnectType, message: str | None = None):
        self.type: DisconnectType = type
        self.message = message

    def to_json(self) -> DisconnectPacketData:
        return {
            "type": self.type.value,
            "message": self.message,
        }

    @classmethod
    def from_json(cls, json: DisconnectPacketData) -> DisconnectPacket:
        return cls(
            type=DisconnectType(json["type"]),
            message=json["message"],
        )


IDENTIFIER = Identifier("core", "packet")


class PACKET_TYPES:
    CONNECT = PacketType.create_json(
        IDENTIFIER,
        "connect",
        Serializer.model(ConnectPacket),
    )
    DISCONNECT = PacketType.create_json(
        IDENTIFIER,
        "disconnect",
        Serializer.model(DisconnectPacket),
    )
    TOKEN = PacketType[str | None].create_json(
        IDENTIFIER,
        "token",
    )
    READY = PacketType[None].create_json(
        IDENTIFIER,
        "ready",
    )



# type generated from https://warehouse.pypa.io/api-reference/json.html
from typing import Literal, TypedDict

type UnknownField = None


class Downloads(TypedDict):
    last_day: int
    last_month: int
    last_week: int


ProjectUrls = TypedDict(
    "ProjectUrls",
    {
        "Bug Reports": str,
        "Funding": str,
        "Homepage": str,
        "Say Thanks!": str,
        "Source": str,
    },
)


class Info(TypedDict):
    author: str
    author_email: str
    bugtrack_url: str | None
    classifiers: list[str]
    description: str
    description_content_type: Literal["text/markdown"]
    docs_url: str | None
    download_url: str
    downloads: Downloads
    home_page: str
    keywords: str
    license: str
    maintainer: str
    maintainer_email: str
    name: str
    package_url: str
    platform: UnknownField
    project_url: str
    project_urls: ProjectUrls
    release_url: str
    requires_dist: list[str]
    requires_python: str
    summary: str
    version: str
    yanked: bool
    yanked_reason: str | None


class Digests(TypedDict):
    blake2b_256: str
    md5: str
    sha256: str


class ReleasesItem(TypedDict):
    comment_text: str
    digests: Digests
    downloads: str
    filename: str
    has_sig: bool
    md5_digest: str
    packagetype: str
    python_version: str
    requires_python: str | None
    size: int
    upload_time: str
    upload_time_iso_8601: str
    url: str
    yanked: bool
    yanked_reason: str | None


type Releases = dict[str, list[ReleasesItem]]
type Urls = list[ReleasesItem]


class VulnerabilitiesItem(TypedDict):
    aliases: list[str]
    details: str
    summary: str
    fixed_in: list[str]
    id: str
    link: str
    source: str
    withdrawn: str | None


type Vulnerabilities = list[VulnerabilitiesItem]


class PackageInfo(TypedDict):
    info: Info
    last_serial: int
    releases: Releases
    urls: Urls
    vulnerabilities: Vulnerabilities



from collections.abc import Callable
from dataclasses import dataclass

from omuserver.server import Server

from omu.helper import Coro

from .client import Client


@dataclass(frozen=True, slots=True)
class Plugin:
    get_client: Callable[[], Client] | None = None
    on_start_server: Coro[[Server], None] | None = None
    on_stop_server: Coro[[Server], None] | None = None
    isolated: bool = False

    def __post_init__(self):
        if self.isolated:
            assert self.on_start_server is None, "Isolated plugins cannot have on_start"
            assert self.on_stop_server is None, "Isolated plugins cannot have on_stop"
            assert self.get_client is not None, "Isolated plugins must have get_client"
